/* ipuz-crossword.c
 *
 * Copyright 2021 Jonathan Blandford <jrb@gnome.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */


#include "libipuz-config.h"
#include "libipuz.h"
#include "libipuz-enums.h"
#include "ipuz-magic.h"
#include <libipuz/ipuz-puzzle.h>
#include <libipuz/ipuz-crossword.h>
#include <libipuz/ipuz-board.h>
#include <glib/gi18n-lib.h>

enum
{
  PROP_0,
  PROP_DIMENSIONS,
  PROP_WIDTH,
  PROP_HEIGHT,
  PROP_BOARD,
  PROP_GUESSES,
  PROP_ZONES,
  PROP_ACROSS_CLUES,
  PROP_DOWN_CLUES,
  PROP_SHOWENUMERATIONS,
  PROP_CLUEPLACEMENT,
  PROP_ANSWER,
  PROP_ANSWERS,
  PROP_ENUMERATION,
  PROP_ENUMERATIONS,
  PROP_MISSES,
  N_PROPERTIES
};
static GParamSpec *obj_props[N_PROPERTIES] = { NULL, };

struct _IPuzCrossword
{
  IPuzPuzzle parent_instance;

  gint width;
  gint height;
  gboolean showenumerations;
  GArray *across_clues;
  GArray *down_clues;

  IPuzBoard *board;
  IPuzGuesses *guesses;

  gboolean uses_extensions;
  gboolean has_solution;
  gboolean has_saved;
  IPuzSymmetry symmetry;
  /*
"zones": [ GroupSpec, ... ]
		Arbitrarily-shaped entries overlaid on the grid
"clueplacement": "before"/"after"/null
		Where to put clues (null = auto)
"clueplacement": "blocks"
		Put clues in blocks adjacent to entry
"answer": "entry"
		The final answer to the puzzle
"answers": [ "entry", ... ]
		List of final answers to the puzzle
"enumeration": Enumeration
		Enumeration of the final answer to the puzzle
"enumerations": [ Enumeration, ... ]
		List of enumerations for final answers to the puzzle
"misses": { "entry": "hint", ... }
		List of hints to be given for misses
   * */
};


static void            ipuz_crossword_init           (IPuzCrossword      *self);
static void            ipuz_crossword_class_init     (IPuzCrosswordClass *klass);
static void            ipuz_crossword_finalize       (GObject            *object);
static void            ipuz_crossword_set_property   (GObject            *object,
                                                      guint               prop_id,
                                                      const GValue       *value,
                                                      GParamSpec         *pspec);
static void            ipuz_crossword_get_property   (GObject            *object,
                                                      guint               prop_id,
                                                      GValue             *value,
                                                      GParamSpec         *pspec);
static void            ipuz_crossword_set_property   (GObject            *object,
                                                      guint               prop_id,
                                                      const GValue       *value,
                                                      GParamSpec         *pspec);
static void            ipuz_crossword_get_property   (GObject            *object,
                                                      guint               prop_id,
                                                      GValue             *value,
                                                      GParamSpec         *pspec);
static void            ipuz_crossword_load_node      (IPuzPuzzle         *puzzle,
                                                      const char         *member_name,
                                                      JsonNode           *node);
static void            ipuz_crossword_post_load_node (IPuzPuzzle         *puzzle,
                                                      const char         *member_name,
                                                      JsonNode           *node);
static void            ipuz_crossword_fixup          (IPuzPuzzle         *puzzle);
static IPuzPuzzleFlags ipuz_crossword_get_flags      (IPuzPuzzle         *puzzle);
static void            ipuz_crossword_build          (IPuzPuzzle         *puzzle,
                                                      JsonBuilder        *builder);
static gboolean        ipuz_crossword_real_set_size  (IPuzCrossword      *xword,
                                                      gint                width,
                                                      gint                height);
static IPuzSymmetry    ipuz_crossword_calc_symmetry  (IPuzCrossword      *xword);



G_DEFINE_TYPE (IPuzCrossword, ipuz_crossword, IPUZ_TYPE_PUZZLE);


/* This is called from GArray, which gives us the address of one of its
 * elements to clear.  Since we are storing (IPuzClue *) in each element,
 * we need double indirection here.
 */
static void
free_one_clue (IPuzClue **clue_ptr)
{
  ipuz_clue_free (*clue_ptr);
  *clue_ptr = NULL;
}

static void
ipuz_crossword_init_clues (IPuzCrossword *self)
{
  self->across_clues = g_array_new (FALSE, TRUE, sizeof (IPuzClue *));
  g_array_set_clear_func (self->across_clues, (GDestroyNotify) free_one_clue);
  self->down_clues = g_array_new (FALSE, TRUE, sizeof (IPuzClue *));
  g_array_set_clear_func (self->down_clues, (GDestroyNotify) free_one_clue);
}

static void
ipuz_crossword_init (IPuzCrossword *self)
{

  g_return_if_fail (self != NULL);

  self->showenumerations = FALSE;

  self->width = 0;
  self->height = 0;
  self->board = ipuz_board_new ();
  ipuz_crossword_init_clues (self);
}

static void
ipuz_crossword_class_init (IPuzCrosswordClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS(klass);
  IPuzPuzzleClass *puzzle_class = IPUZ_PUZZLE_CLASS (klass);

  object_class->finalize = ipuz_crossword_finalize;
  object_class->set_property = ipuz_crossword_set_property;
  object_class->get_property = ipuz_crossword_get_property;
  puzzle_class->load_node = ipuz_crossword_load_node;
  puzzle_class->post_load_node = ipuz_crossword_post_load_node;
  puzzle_class->fixup = ipuz_crossword_fixup;
  puzzle_class->get_flags = ipuz_crossword_get_flags;
  puzzle_class->build = ipuz_crossword_build;

  obj_props[PROP_WIDTH] = g_param_spec_int ("width",
                                            _("width"),
                                            _("Width of the puzzle grid"),
                                            0, 65536, 0,
                                            G_PARAM_READWRITE);
  g_object_class_install_property (object_class, PROP_WIDTH, obj_props[PROP_WIDTH]);

  obj_props[PROP_HEIGHT] = g_param_spec_int ("height",
                                             _("height"),
                                             _("height of the puzzle grid"),
                                             0, 65536, 0,
                                             G_PARAM_READWRITE);
  g_object_class_install_property (object_class, PROP_HEIGHT, obj_props[PROP_HEIGHT]);

  obj_props[PROP_SHOWENUMERATIONS] = g_param_spec_boolean ("showenumerations",
                                                           _("Show Enumerations"),
                                                           _("Show enumerations with clues"),
                                                           FALSE,
                                                           G_PARAM_READWRITE);
  g_object_class_install_property (object_class, PROP_SHOWENUMERATIONS, obj_props[PROP_SHOWENUMERATIONS]);

  obj_props[PROP_BOARD] = g_param_spec_object ("board",
                                               _("Board"),
                                               _("The crossword board"),
                                               IPUZ_TYPE_BOARD,
                                               G_PARAM_READABLE);
  g_object_class_install_property (object_class, PROP_BOARD, obj_props[PROP_BOARD]);

  obj_props[PROP_GUESSES] = g_param_spec_boxed ("guesses",
                                                _("Guesses"),
                                                _("The guesses associated with this crossword"),
                                                IPUZ_TYPE_GUESSES,
                                                G_PARAM_READWRITE);
  g_object_class_install_property (object_class, PROP_GUESSES, obj_props[PROP_GUESSES]);
}

/* free all memory */
static void
ipuz_crossword_finalize (GObject *object)
{
  IPuzCrossword *self;

  g_return_if_fail (object != NULL);

  self = IPUZ_CROSSWORD (object);

  g_clear_pointer (&self->across_clues, g_array_unref);
  g_clear_pointer (&self->down_clues, g_array_unref);

  g_object_unref (G_OBJECT (self->board));
  ipuz_guesses_unref (self->guesses);

  G_OBJECT_CLASS (ipuz_crossword_parent_class)->finalize (object);
}

static void
ipuz_crossword_set_property (GObject      *object,
                             guint         prop_id,
                             const GValue *value,
                             GParamSpec   *pspec)
{
  IPuzCrossword *self;

  g_return_if_fail (object != NULL);

  self = IPUZ_CROSSWORD (object);

  switch (prop_id)
    {
    case PROP_WIDTH:
      ipuz_crossword_real_set_size (self,
                                    g_value_get_int (value),
                                    self->height);
      break;
    case PROP_HEIGHT:
      ipuz_crossword_real_set_size (self,
                                    self->width,
                                    g_value_get_int (value));
      break;
    case PROP_SHOWENUMERATIONS:
      self->showenumerations = g_value_get_boolean (value);
      break;
    case PROP_GUESSES:
      ipuz_crossword_set_guesses (self, (IPuzGuesses *)g_value_get_boxed (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
ipuz_crossword_get_property (GObject *object,
                             guint prop_id,
                             GValue *value,
                             GParamSpec *pspec)
{
  IPuzCrossword *self;

  g_return_if_fail (G_IS_OBJECT (object));

  self = IPUZ_CROSSWORD (object);

  switch (prop_id)
    {
    case PROP_WIDTH:
      g_value_set_int (value, self->width);
      break;
    case PROP_HEIGHT:
      g_value_set_int (value, self->height);
      break;
    case PROP_SHOWENUMERATIONS:
      g_value_set_boolean (value, self->showenumerations);
      break;
    case PROP_BOARD:
      g_value_set_object (value, self->board);
      break;
    case PROP_GUESSES:
      g_value_set_boxed (value, self->guesses);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
load_dimensions (IPuzCrossword *xword,
                 JsonNode      *node)
{
  gint width = -1;
  gint height = -1;

  if (!JSON_NODE_HOLDS_OBJECT (node))
    return;

  g_autoptr(JsonReader) reader = json_reader_new (node);
  if (json_reader_read_member (reader, "width"))
    width = json_reader_get_int_value (reader);
  json_reader_end_member (reader);

  if (json_reader_read_member (reader, "height"))
    height = json_reader_get_int_value (reader);
  json_reader_end_member (reader);

  if (width >= 0 && height >= 0)
    ipuz_crossword_real_set_size (xword, width, height);
}

static void
load_clues_foreach (JsonArray *array,
                    guint      index,
                    JsonNode  *element_node,
                    gpointer   user_data)
{
  IPuzClue *clue;

  clue = ipuz_clue_new_from_json (element_node);
  GArray *clues = (GArray *) user_data;
  g_array_append_val (clues, clue);
}

static void
load_clues (IPuzCrossword *self,
            JsonNode      *node)
{
  JsonObjectIter iter = {0, };
  const gchar *member_name = NULL;
  JsonNode *member_node;

  if (!JSON_NODE_HOLDS_OBJECT (node))
    return;

  json_object_iter_init (&iter, json_node_get_object (node));
  while (json_object_iter_next (&iter, &member_name, &member_node))
    {
      GArray *clues = NULL;
      IPuzClueDirection direction;
      /* conditions for parsing clues. It has to be a direction we know, and has
       * to be an array.
       */
      if (! JSON_NODE_HOLDS_ARRAY (member_node))
        continue;

      if (g_strcmp0 ("Across", member_name) == 0)
        {
          clues = self->across_clues;
          direction = IPUZ_CLUE_DIRECTION_ACROSS;
        }
      else if (g_strcmp0 ("Down", member_name) == 0)
        {
          clues = self->down_clues;
          direction = IPUZ_CLUE_DIRECTION_DOWN;
        }
      else
        continue;

      g_assert (clues != NULL);

      if (clues->len > 0)
        g_array_set_size (clues, 0);

      json_array_foreach_element (json_node_get_array (member_node), load_clues_foreach, clues);

      /* The json section of each clue doesn't contain the direction. Go back and set it
       * as a second pass.
       */
      for (guint i = 0; i < clues->len; i++)
        ipuz_clue_set_direction (g_array_index (clues, IPuzClue *, i), direction);
    }
}

static void
ipuz_crossword_load_node (IPuzPuzzle *puzzle,
                          const char *member_name,
                          JsonNode   *node)
{

  g_return_if_fail (IPUZ_IS_PUZZLE (puzzle));
  g_return_if_fail (member_name != NULL);
  g_return_if_fail (node != NULL);

  if (strcmp (member_name, "dimensions") == 0)
    {
      load_dimensions (IPUZ_CROSSWORD (puzzle), node);
      return;
    }
  if (strcmp (member_name, "clues") == 0)
    {
      load_clues (IPUZ_CROSSWORD (puzzle), node);
      return;
    }

  IPUZ_PUZZLE_CLASS (ipuz_crossword_parent_class)->load_node (puzzle, member_name, node);
}


static void
ipuz_crossword_post_load_node (IPuzPuzzle *puzzle,
                               const char *member_name,
                               JsonNode   *node)
{
  IPuzCrossword *self;

  g_return_if_fail (IPUZ_IS_PUZZLE (puzzle));
  g_return_if_fail (member_name != NULL);
  g_return_if_fail (node != NULL);

  self = IPUZ_CROSSWORD (puzzle);

  if (strcmp (member_name, "puzzle") == 0)
    {
      g_autofree gchar *block, *empty;
      g_object_get (G_OBJECT (puzzle),
                    "block", &block,
                    "empty", &empty,
                    NULL);
      ipuz_board_parse_puzzle (self->board, node, block, empty);
      return;
    }
  if (strcmp (member_name, "solution") == 0)
    {
      g_autofree gchar *block, *charset;
      g_object_get (G_OBJECT (puzzle),
                    "block", &block,
                    "charset", &charset,
                    NULL);
      ipuz_board_parse_solution (self->board, node, block, charset);
      self->has_solution = TRUE;
      return;
    }
  if (strcmp (member_name, "saved") == 0)
    {
      /* FIXME(saved): Load the saved guesses from disk */
    }
  /* We do not chain up! */
}

/* This function makes sure that each clue has its "cells" set correctly.
 * Either its explicitly set by the puzzle, or implicitly done by the board
 * itself. Either way, we need to link back from the Cell to the Clue.
 *
 * Question: Should we move this to IPuzClue? We should also split it in half.
 * */
static void
crossword_ensure_cells (IPuzClue  *clue,
                        IPuzBoard *board)
{
  const GArray *cells;
  IPuzCellCoord coord;
  IPuzCell *cell;

  cells = ipuz_clue_get_cells (clue);

  /* the clue already has cells loaded. */
  if (cells->len > 0)
    {
      for (guint i = 0; i < cells->len; i ++)
        {
          coord = g_array_index (clue->cells, IPuzCellCoord, i);
          cell = ipuz_board_get_cell (board, coord);
          ipuz_cell_set_clue (cell, clue);
        }
      return;
    }

  cell = ipuz_board_get_cell_by_clue (board, clue, &coord);

  if (cell == NULL)
    /* We have a clue that's not in the grid. We'll flag this in validate stage
     * but will ahve to keep its cells empty now.
     */
    return;

  while (cell != NULL && ipuz_cell_get_cell_type (cell) == IPUZ_CELL_NORMAL)
    {
      ipuz_clue_append_cell (clue, coord);
      ipuz_cell_set_clue (cell, clue);

      switch (ipuz_clue_get_direction (clue))
        {
        case IPUZ_CLUE_DIRECTION_ACROSS:
          coord.column ++;
          break;
        case IPUZ_CLUE_DIRECTION_DOWN:
          coord.row ++;
          break;
        default:
          g_assert_not_reached ();
        }
      cell = ipuz_board_get_cell (board, coord);
    }
}

static void
ipuz_crossword_fixup_cells (IPuzCrossword *self)
{
  g_autoptr(GHashTable) styles = NULL;
  guint row, col;

  g_object_get (G_OBJECT (self),
                "styles", &styles,
                NULL);

  if (styles == NULL)
    return;

  for (row = 0; row < (guint) self->height; row++)
    {
      for (col = 0; col < (guint) self->width; col++)
        {
          IPuzCell *cell;
          IPuzCellCoord coord = { .row = row, .column = col };

          cell = ipuz_crossword_get_cell (self, coord);
          if (cell->style_name)
            {
              IPuzStyle *style;

              style = g_hash_table_lookup (styles, cell->style_name);
              if (style)
                {
                  ipuz_style_ref (style);
                  g_clear_pointer (&cell->style, ipuz_style_unref);
                  cell->style = style;
                }
            }

          if (IPUZ_CELL_IS_BLOCK (cell) && cell->style != NULL)
            self->uses_extensions = TRUE;

          if (cell->saved_guess != NULL)
            self->has_saved = TRUE;
        }
    }

}


/* Make sure every clue has its "cells" set.
 * Update the enumeration if necessary
 */
static void
ipuz_crossword_fixup_clues (IPuzCrossword *self)
{
  IPuzClue *clue;
  guint i;

  for (i = 0; i < self->across_clues->len; i++)
    {
      clue = g_array_index (self->across_clues, IPuzClue *, i);
      crossword_ensure_cells (clue, self->board);
      if (self->showenumerations)
        ipuz_clue_ensure_enumeration (clue);
    }

  for (i = 0; i < self->down_clues->len; i++)
    {
      clue = g_array_index (self->down_clues, IPuzClue *, i);
      crossword_ensure_cells (clue, self->board);
      if (self->showenumerations)
        ipuz_clue_ensure_enumeration (clue);
    }
}


static void
ipuz_crossword_fixup (IPuzPuzzle *puzzle)
{
  IPuzCrossword *self;

  g_return_if_fail (IPUZ_IS_CROSSWORD (puzzle));

  self = IPUZ_CROSSWORD (puzzle);

  ipuz_crossword_fixup_clues (self);
  ipuz_crossword_fixup_cells (self);

  self->symmetry = ipuz_crossword_calc_symmetry (self);

  IPUZ_PUZZLE_CLASS (ipuz_crossword_parent_class)->fixup (puzzle);
}


static IPuzPuzzleFlags
ipuz_crossword_get_flags (IPuzPuzzle *puzzle)
{
  IPuzCrossword *self;
  guint flags;

  g_return_val_if_fail (IPUZ_IS_CROSSWORD (puzzle), 0);

  self = IPUZ_CROSSWORD (puzzle);

  flags = IPUZ_PUZZLE_CLASS (ipuz_crossword_parent_class)->get_flags (puzzle);

  if (self->across_clues->len != 0 || self->down_clues->len != 0)
    flags |= IPUZ_PUZZLE_FLAG_HAS_CLUES;


  if (self->uses_extensions)
    flags |= IPUZ_PUZZLE_FLAG_USES_EXTENSIONS;

  if (self->has_solution)
    flags |= IPUZ_PUZZLE_FLAG_HAS_SOLUTION;

  if (self->has_saved)
    flags |= IPUZ_PUZZLE_FLAG_HAS_SAVED;

  return flags;
}

static void
build_dimensions (IPuzCrossword *xword,
                  JsonBuilder   *builder)
{
  json_builder_set_member_name (builder, "dimensions");
  json_builder_begin_object (builder);
  json_builder_set_member_name (builder, "width");
  json_builder_add_int_value (builder, xword->width);
  json_builder_set_member_name (builder, "height");
  json_builder_add_int_value (builder, xword->height);
  json_builder_end_object (builder);
}

static void
build_clues (IPuzCrossword *xword,
             JsonBuilder   *builder,
             const gchar   *member,
             GArray        *clues)
{
  guint i;
  if (clues->len == 0)
    return;
  json_builder_set_member_name (builder, member);
  json_builder_begin_array (builder);
  for (i = 0; i < clues->len; i++)
    {
      IPuzClue *clue;
      clue = g_array_index (clues, IPuzClue *, i);
      ipuz_clue_build (clue, builder);
    }
  json_builder_end_array (builder);
}

static void
ipuz_crossword_build (IPuzPuzzle  *puzzle,
                      JsonBuilder *builder)
{
  IPuzCrossword *xword;
  g_autofree gchar *block = NULL;
  g_autofree gchar *empty = NULL;

  g_return_if_fail (IPUZ_IS_CROSSWORD (puzzle));

  xword = IPUZ_CROSSWORD (puzzle);
  g_object_get (G_OBJECT (puzzle),
                "block", &block,
                "empty", &empty,
                NULL);

  /* We chain to the parent class first. Not every implementation can
   * handle meta-information at the end. */
  IPUZ_PUZZLE_CLASS (ipuz_crossword_parent_class)->build (puzzle, builder);

  build_dimensions (xword, builder);
  json_builder_set_member_name (builder, "showenumerations");
  json_builder_add_boolean_value (builder, xword->showenumerations);


  ipuz_board_build_puzzle (xword->board, builder, block, empty);
  ipuz_board_build_solution (xword->board, builder, block);

  if (xword->across_clues->len > 0 ||
      xword->down_clues->len > 0)
    {
      json_builder_set_member_name (builder, "clues");
      json_builder_begin_object (builder);
      build_clues (xword, builder, "Across", xword->across_clues);
      build_clues (xword, builder, "Down", xword->down_clues);
      json_builder_end_object (builder);
    }
}

/**
 *
 */
IPuzCrossword *
ipuz_crossword_new ()
{
  IPuzCrossword *xword;

  xword = IPUZ_CROSSWORD (g_object_new (IPUZ_TYPE_CROSSWORD,
                                        "puzzle-kind", IPUZ_PUZZLE_CROSSWORD,
                                        "version", "http://ipuz.org/v2",
                                        NULL));

  return xword;
}

/*
 * Copying boards is painful. They don't exist on their own but refer
 * to other portions of the crossword. There's not really a good way to
 * just make fresh copies of cells, as they refer to styles and clues
 * within the puzzle. We can copy the board and cells on the surface,
 * then refix them up afterwards. Regardless of which, Boards don't
 * exist outside the context of their Crossword in practice.
 */
static void
ipuz_crossword_deep_clone_board (IPuzCrossword *src,
                                 IPuzCrossword *dest)
{
  GHashTable *styles = NULL;
  guint row, column;

  ipuz_board_resize (dest->board, dest->width, dest->height);
  g_object_get (G_OBJECT (dest),
                "styles", &styles,
                NULL);

  for (row = 0; row < (guint) dest->height; row++)
    {
      for (column = 0; column < (guint) dest->width; column++)
        {
          IPuzCellCoord coord = { .row = row, .column = column };
          IPuzCell *src_cell, *dest_cell;
          IPuzStyle *style;

          src_cell = ipuz_crossword_get_cell (src, coord);
          dest_cell = ipuz_crossword_get_cell (dest, coord);

          dest_cell->cell_type = src_cell->cell_type;
          dest_cell->number = src_cell->number;
          dest_cell->label = g_strdup (src_cell->label);
          dest_cell->solution = g_strdup (src_cell->solution);
          dest_cell->initial_val = g_strdup (src_cell->initial_val);
          dest_cell->saved_guess = g_strdup (src_cell->saved_guess);
          dest_cell->style_name = g_strdup (src_cell->style_name);

          if (dest_cell->style_name && styles)
            {
              style = g_hash_table_lookup (styles, dest_cell->style_name);
              if (style)
                dest_cell->style = ipuz_style_ref (style);
            }

          /* update the clues. We will have already copied the clues before we attempt to do this */
          if (src_cell->across_clue)
            {
              guint clue_num;

              clue_num = ipuz_crossword_find_clue (src, src_cell->across_clue);
              dest_cell->across_clue = ipuz_crossword_get_clue (dest, IPUZ_CLUE_DIRECTION_ACROSS, clue_num);
            }

          if (src_cell->down_clue)
            {
              guint clue_num;

              clue_num = ipuz_crossword_find_clue (src, src_cell->down_clue);
              dest_cell->down_clue = ipuz_crossword_get_clue (dest, IPUZ_CLUE_DIRECTION_DOWN, clue_num);
            }

          /* There's a chance the original crossword has a weird state
           * with a style_name and an independent style. We keep that
           *i nconsistent state rather than trying to clean it up. */
          if (dest_cell->style == NULL)
            dest_cell->style = ipuz_style_copy (src_cell->style);
        }
    }
}

IPuzCrossword *
ipuz_crossword_deep_copy (IPuzCrossword *src)
{
  IPuzCrossword *dest;
  guint i;

  if (src == NULL)
    return NULL;

  dest = ipuz_crossword_new ();

  /* We should make this a virtual function */
  _ipuz_puzzle_deep_clone (IPUZ_PUZZLE (src), IPUZ_PUZZLE (dest));

  dest->width = src->width;
  dest->height = src->height;
  dest->showenumerations = src->showenumerations;

  for (i = 0; i < src->across_clues->len; i++)
    {
      IPuzClue *clue = g_array_index (src->across_clues, IPuzClue *, i);
      IPuzClue *new_clue = ipuz_clue_copy (clue);
      g_array_append_val (dest->across_clues, new_clue);
    }

  for (i = 0; i < src->down_clues->len; i++)
    {
      IPuzClue *clue = g_array_index (src->down_clues, IPuzClue *, i);
      IPuzClue *new_clue = ipuz_clue_copy (clue);
      g_array_append_val (dest->down_clues, new_clue);
    }

  ipuz_crossword_deep_clone_board (src, dest);
  dest->guesses = ipuz_guesses_copy (src->guesses);

  dest->uses_extensions = src->uses_extensions;
  dest->has_solution = src->has_solution;
  dest->has_saved = src->has_saved;
  dest->symmetry = src->symmetry;

  return dest;
}


guint
ipuz_crossword_get_width (IPuzCrossword *xword)
{
  g_return_val_if_fail (IPUZ_IS_CROSSWORD (xword), 0);

  return xword->width;
}


guint
ipuz_crossword_get_height (IPuzCrossword *xword)
{
  g_return_val_if_fail (IPUZ_IS_CROSSWORD (xword), 0);

  return xword->height;
}


static gboolean
ipuz_crossword_real_set_size (IPuzCrossword *xword,
                              gint           width,
                              gint           height)
{
  if (xword->width == width && xword->height == height)
    return FALSE;

  xword->width = width;
  xword->height = height;

  ipuz_board_resize (xword->board, width, height);

  return TRUE;
}


void
ipuz_crossword_set_size   (IPuzCrossword *xword,
                           gint           width,
                           gint           height)
{
 g_return_if_fail (IPUZ_IS_CROSSWORD (xword));

  g_object_freeze_notify (G_OBJECT (xword));
  if (ipuz_crossword_real_set_size (xword, width, height))
    {
      g_object_notify_by_pspec (G_OBJECT (xword), obj_props[PROP_WIDTH]);
      g_object_notify_by_pspec (G_OBJECT (xword), obj_props[PROP_HEIGHT]);
    }
  g_object_thaw_notify (G_OBJECT (xword));
}

IPuzBoard *
ipuz_crossword_get_board (IPuzCrossword *xword)
{
  g_return_val_if_fail (IPUZ_IS_CROSSWORD (xword), NULL);

  return xword->board;
}

static gboolean
ipuz_crossword_validate_guesses (IPuzCrossword *xword,
                                 IPuzGuesses   *guesses)
{
  guint row, col;

  if ((guint) xword->width != ipuz_guesses_get_width (guesses) ||
      (guint) xword->height != ipuz_guesses_get_height (guesses))
    return FALSE;

  for (row = 0; row < (guint) xword->height; row++)
    {
      for (col = 0; col < (guint) xword->width; col++)
        {
          IPuzCell *cell;
          IPuzCellCellType guess_type;
          IPuzCellCoord coord = { .row = row, .column = col };

          cell = ipuz_crossword_get_cell (xword, coord);
          guess_type = ipuz_guesses_get_cell_type (guesses, coord);

          /* We check to make sure that if the cell is guessable, we
           * can write to it. We ignore the remainder of the cells. */
          if (IPUZ_CELL_IS_GUESSABLE (cell) &&
              guess_type != IPUZ_CELL_NORMAL)
            return FALSE;
        }
    }
  return TRUE;
}

gboolean
ipuz_crossword_set_guesses (IPuzCrossword *xword,
                            IPuzGuesses   *guesses)
{
  g_return_val_if_fail (IPUZ_IS_CROSSWORD (xword), FALSE);

  if (guesses)
    {
      if (! ipuz_crossword_validate_guesses (xword, guesses))
        return FALSE;
      ipuz_guesses_ref (guesses);
    }

  g_clear_pointer (&xword->guesses, ipuz_guesses_unref);
  xword->guesses = guesses;
  g_object_notify_by_pspec (G_OBJECT (xword), obj_props[PROP_GUESSES]);

  return TRUE;
}

IPuzGuesses *
ipuz_crossword_get_guesses (IPuzCrossword *xword)
{
  g_return_val_if_fail (IPUZ_IS_CROSSWORD (xword), NULL);

  return xword->guesses;
}

/* Convenience function. */
IPuzCell *
ipuz_crossword_get_cell (IPuzCrossword *xword,
                         IPuzCellCoord  coord)
{
  g_return_val_if_fail (IPUZ_IS_CROSSWORD (xword), NULL);

  g_return_val_if_fail (xword->board != NULL, NULL);

  return ipuz_board_get_cell (xword->board, coord);
}


GArray *
ipuz_crossword_get_clues (IPuzCrossword     *xword,
                          IPuzClueDirection  direction)
{
  g_return_val_if_fail (IPUZ_IS_CROSSWORD (xword), NULL);
  g_return_val_if_fail (direction == IPUZ_CLUE_DIRECTION_ACROSS ||
                        direction == IPUZ_CLUE_DIRECTION_DOWN, NULL);

  if (direction == IPUZ_CLUE_DIRECTION_ACROSS)
    return xword->across_clues;
  else
    return xword->down_clues;
}

IPuzClue *
ipuz_crossword_get_clue (IPuzCrossword     *xword,
                         IPuzClueDirection  direction,
                         guint              index)
{
  GArray *clues;

  g_return_val_if_fail (IPUZ_IS_CROSSWORD (xword), NULL);

  clues = ipuz_crossword_get_clues (xword, direction);

  return g_array_index (clues, IPuzClue *, index);
}


guint
ipuz_crossword_get_n_clues (IPuzCrossword     *xword,
                            IPuzClueDirection  direction)
{
  GArray *clues = NULL;

  g_return_val_if_fail (IPUZ_IS_CROSSWORD (xword), 0);
  g_return_val_if_fail (direction == IPUZ_CLUE_DIRECTION_ACROSS ||
                        direction == IPUZ_CLUE_DIRECTION_DOWN, 0);

  if (direction == IPUZ_CLUE_DIRECTION_ACROSS)
    clues = xword->across_clues;
  else
    clues = xword->down_clues;

  return clues->len;
}

guint
ipuz_crossword_find_clue (IPuzCrossword *xword,
                          IPuzClue      *clue)
{
  GArray *clues = NULL;

  g_return_val_if_fail (IPUZ_IS_CROSSWORD (xword), 0);
  g_return_val_if_fail (clue != NULL, 0);

  clues = ipuz_crossword_get_clues (xword, clue->direction);

  for (guint i = 0; i < clues->len; i ++)
    {
      IPuzClue *tmp_clue = g_array_index (clues, IPuzClue *, i);
      if (ipuz_clue_compare (clue, tmp_clue))
        return i;
    }
  return 0;

}

gchar *
ipuz_crossword_get_string (IPuzCrossword     *xword,
                           IPuzClueDirection  direction,
                           gint               number,
                           gboolean           guesses)
{
  IPuzClue *clue;
  const GArray *cells;
  GString *string;
  guint i;

  g_return_val_if_fail (IPUZ_IS_CROSSWORD (xword), NULL);

  clue = ipuz_crossword_get_clue (xword, direction, number);

  if (clue == NULL)
    return NULL;

  string = g_string_new (NULL);
  cells = ipuz_clue_get_cells (clue);
  for (i = 0; i < cells->len; i++)
    {
      IPuzCellCoord coord;
      const char *solution = NULL;

      coord = g_array_index (cells, IPuzCellCoord, i);
      if (guesses)
        {
          solution = ipuz_guesses_get_guess (xword->guesses, coord);
        }
      else
        {
          IPuzCell *cell;

          cell = ipuz_crossword_get_cell (xword, coord);
          solution = ipuz_cell_get_solution (cell);
        }

      if (solution && solution[0] !='\0')
        g_string_append (string, solution);
      else
        g_string_append (string, "?");
    }

  return g_string_free (string, FALSE);
}


/**
 * ipuz_crossword_get_clue_string:
 * @xword: An 'IPuzCrossword'
 * @direction: Direction of the clue
 * @number: index of the clue
 *
 * Returns a string containing the solution of the puzzle for a given
 * clue. This string will have '?' characterd embedded within it if
 * there are cells without solutions set yet.
 *
 * Returns: The solution for the puzzle for Clue, or %NULL
 **/
gchar *
ipuz_crossword_get_clue_string (IPuzCrossword     *xword,
                                IPuzClueDirection  direction,
                                gint               number)
{
  g_return_val_if_fail (IPUZ_IS_CROSSWORD (xword), NULL);

  return ipuz_crossword_get_string (xword, direction, number, FALSE);
}

/**
 * ipuz_crossword_get_guess_string:
 * @xword: An 'IPuzCrossword'
 * @direction: Direction of the clue
 * @number: index of the clue
 *
 * Returns a string containing the guess in the puzzle for a given
 * clue. This string will have '?' characterd embedded within it if
 * there are cells not completely filled out.
 *
 * Returns: The solution for the puzzle for Clue, or %NULL
 **/
gchar *
ipuz_crossword_get_guess_string (IPuzCrossword     *xword,
                                IPuzClueDirection  direction,
                                gint               number)
{
  g_return_val_if_fail (IPUZ_IS_CROSSWORD (xword), NULL);
  g_return_val_if_fail (xword->guesses != NULL, NULL);

  return ipuz_crossword_get_string (xword, direction, number, TRUE);
}

/**
 * ipuz_crossword_clue_guessed:
 * @xword:  An 'IPuzCrossword'
 * @clue: The clue to check
 *
 * Check to see if every field in the clue is guessed. This doesn't
 * account for the correctness of the clue
 *
 * Returns: %TRUE, if every field in the clue is guessed.
 **/
gboolean
ipuz_crossword_clue_guessed (IPuzCrossword *xword,
                             IPuzClue      *clue,
                             gboolean      *correct)
{
  const GArray *cells;
  guint i;
  gboolean guessed = TRUE;

  g_return_val_if_fail (IPUZ_IS_CROSSWORD (xword), FALSE);
  g_return_val_if_fail (clue != NULL, FALSE);

  cells = ipuz_clue_get_cells (clue);
  if (cells == NULL)
    return FALSE;

  if (correct)
    *correct = TRUE;

  for (i = 0; i < cells->len; i++)
    {
      IPuzCellCoord coord;
      IPuzCell *cell;
      const char *guess = NULL;
      const char *solution = NULL;

      coord = g_array_index (cells, IPuzCellCoord, i);
      cell = ipuz_crossword_get_cell (xword, coord);

      guess = ipuz_guesses_get_guess (xword->guesses, coord);
      solution = ipuz_cell_get_solution (cell);

      if (guess == NULL || guess[0] == '\0')
        guessed = FALSE;
      if (correct)
        *correct = !g_strcmp0 (solution, guess) && *correct;
    }

  return guessed;
}

IPuzClue *
ipuz_crossword_find_clue_by_number (IPuzCrossword     *xword,
                                    IPuzClueDirection  direction,
                                    gint               number)
{
  GArray *clues = NULL;

  g_return_val_if_fail (IPUZ_IS_CROSSWORD (xword), NULL);
  g_return_val_if_fail (direction == IPUZ_CLUE_DIRECTION_ACROSS ||
                        direction == IPUZ_CLUE_DIRECTION_DOWN, NULL);

  clues = ipuz_crossword_get_clues (xword, direction);

  for (guint i = 0; i < clues->len; i ++)
    {
      IPuzClue *clue;
      clue = g_array_index (clues, IPuzClue *, i);
      if (clue->number == number)
        return clue;
    }
  return NULL;
}

IPuzClue *
ipuz_crossword_find_clue_by_label (IPuzCrossword     *xword,
                                   IPuzClueDirection  direction,
                                   const gchar       *label)
{
  GArray *clues = NULL;

  g_return_val_if_fail (IPUZ_IS_CROSSWORD (xword), NULL);
  g_return_val_if_fail (label != NULL, NULL);
  g_return_val_if_fail (direction == IPUZ_CLUE_DIRECTION_ACROSS ||
                        direction == IPUZ_CLUE_DIRECTION_DOWN, NULL);

  clues = ipuz_crossword_get_clues (xword, direction);

  for (guint i = 0; i < clues->len; i ++)
    {
      IPuzClue *clue;
      clue = g_array_index (clues, IPuzClue *, i);
      if (g_strcmp0 (label, clue->label) == 0)
        return clue;
    }
  return NULL;
}

IPuzClue *
ipuz_crossword_find_clue_by_coord (IPuzCrossword     *xword,
                                   IPuzClueDirection  direction,
                                   IPuzCellCoord      coord)
{
 GArray *clues = NULL;

  g_return_val_if_fail (IPUZ_IS_CROSSWORD (xword), NULL);
  g_return_val_if_fail (direction == IPUZ_CLUE_DIRECTION_ACROSS ||
                        direction == IPUZ_CLUE_DIRECTION_DOWN, NULL);

  clues = ipuz_crossword_get_clues (xword, direction);

  for (guint i = 0; i < clues->len; i ++)
    {
      IPuzClue *clue;
      clue = g_array_index (clues, IPuzClue *, i);
      if (ipuz_clue_contains_cell (clue, coord))
        return clue;
    }
  return NULL;
}

static gboolean
ipuz_crossword_game_won_solution (IPuzCrossword *xword)
{
 guint row, col;

 for (row = 0; row < (guint) xword->height; row++)
   {
     for (col = 0; col < (guint) xword->width; col++)
       {
         IPuzCell *cell;
         IPuzCellCoord coord = { .row = row, .column = col };

         cell = ipuz_crossword_get_cell (xword, coord);
         if (IPUZ_CELL_IS_GUESSABLE (cell))
           {
             const char *guess = ipuz_guesses_get_guess (xword->guesses, coord);
             /* FIXME(charset): There are puzzles where the solution charset
              * doesn't match the solution and we have a bunch of
              * blank solution values. We want to catch that case.
              */
             /* FIXME(charset): should we normalize guess and solution first?*/
             if (guess == NULL)
               return FALSE;
             if (g_strcmp0 (guess, cell->solution))
               return FALSE;
           }
       }
   }

 return TRUE;
}

gboolean
ipuz_crossword_game_won (IPuzCrossword *xword)
{
  guint flags;

  g_return_val_if_fail (IPUZ_IS_CROSSWORD (xword), FALSE);

  /* You can't win without playing! */
  if (xword->guesses == NULL)
    return FALSE;

  flags = ipuz_puzzle_get_flags (IPUZ_PUZZLE (xword));

  if (flags & IPUZ_PUZZLE_FLAG_HAS_SOLUTION)
    return ipuz_crossword_game_won_solution (xword);

  if (flags & IPUZ_PUZZLE_FLAG_HAS_CHECKSUM)
    {
      /* FIXME(checksum): need to support this! */
    }
  return FALSE;
}


static IPuzClue *
calculate_clue (IPuzCrossword     *xword,
                IPuzClueDirection  direction,
                IPuzCellCoord      coord,
                gint               number)
{
  IPuzClue *clue;
  IPuzCell *cell;
  gboolean valid = FALSE;
  guint len = 0;

  /* If we're already in a clue, we don't want to create another one */
  if (ipuz_crossword_find_clue_by_coord (xword, direction, coord) != NULL)
    return NULL;

  /* create a perspective new clue, and see if it's valid. We do this
   * by seeing if we can walk in direction and append the cell to the
   * clue. If our len is > 1, then it means that it's a valid
   * clue. Also, we do a dance between NORMAL cells and GESSABLE cells
   * because we can have a row of all guessable cells that shouldn't
   * be calculated. See the first puzzle as an example of this. */
  clue = ipuz_clue_new ();
  ipuz_clue_set_direction (clue, direction);
  ipuz_clue_set_number (clue, number);

  cell = ipuz_crossword_get_cell (xword, coord);
  while (IPUZ_CELL_IS_NORMAL (cell))
    {
      ipuz_clue_append_cell (clue, coord);
      len ++;
      if (IPUZ_CELL_IS_GUESSABLE (cell))
        valid = TRUE;

      if (direction == IPUZ_CLUE_DIRECTION_ACROSS)
        coord.column ++;
      else
        coord.row ++;
      cell = ipuz_crossword_get_cell (xword, coord);
    }

  if (valid && len > 1)
    return clue;

  ipuz_clue_free (clue);
  return NULL;
}

/**
 * ipuz_crossword_calculate_clues:
 * @xword: An 'IPuzCrossword'
 *
 * Mutates the puzzle and sets valid clues and clue numbers based on
 * BLOCKS. It will clear the existing clues and reset every cell's
 * number in the board. It can be called after the overall structure
 * of the board changes (ie, the number of BLOCKs changes) to
 * regenerate a consistent puzzle.
 **/
void
ipuz_crossword_calculate_clues (IPuzCrossword *xword)
{
  guint row, col;
  gint number = 1;

  g_return_if_fail (IPUZ_IS_CROSSWORD (xword));

  /* Clear all clues. We are starting over. */
  g_clear_pointer (&xword->across_clues, g_array_unref);
  g_clear_pointer (&xword->down_clues, g_array_unref);
  ipuz_crossword_init_clues (xword);

  for (row = 0; row < (guint) xword->height; row++)
    {
      for (col = 0; col < (guint) xword->width; col++)
        {
          IPuzCell *cell;
          IPuzCellCoord coord = { .row = row, .column = col };

          cell = ipuz_crossword_get_cell (xword, coord);
          /* Clear out the old clue */
          cell->across_clue = NULL;
          cell->down_clue = NULL;

          if (IPUZ_CELL_IS_NORMAL (cell))
            {
              IPuzClue *across_clue;
              IPuzClue *down_clue;

              across_clue = calculate_clue (xword, IPUZ_CLUE_DIRECTION_ACROSS, coord, number);
              down_clue = calculate_clue (xword, IPUZ_CLUE_DIRECTION_DOWN, coord, number);

              if (across_clue)
                g_array_append_val (xword->across_clues, across_clue);
              if (down_clue)
                g_array_append_val (xword->down_clues, down_clue);


              if (across_clue || down_clue)
                ipuz_cell_set_number (cell, number++);
              else
                ipuz_cell_set_number (cell, 0);

              /* FIXME: Should we clear labels as well? */
            }
        }
    }
  ipuz_crossword_fixup_clues (xword);
}

/* If offset == 0, then we do the opposite cell,
 * if it's -1, then we do the counter-clockwise adjacent,
 * if it's 1, then we do the clockwise adjacent */

static IPuzCell *
ipuz_crossword_cell_symmetry (IPuzCrossword *xword,
                              IPuzCellCoord  coord,
                              gint           offset)
{
  IPuzCellCoord mirror_coord;

  if (xword->width != xword->height &&
      offset != 0)
    {
      g_warning ("Trying to enfoce Quarter cell-symmetry on a non-square board\n");
      return NULL;
    }

  if (offset == 0)
    {
      mirror_coord.row = xword->height - coord.row - 1;
      mirror_coord.column = xword->width - coord.column - 1;
    }
  else if (offset == -1)
    {
      mirror_coord.row = xword->height - coord.column - 1;
      mirror_coord.column = coord.row;
    }
  else if (offset == 1)
    {
      mirror_coord.row = coord.column;
      mirror_coord.column = xword->width - coord.row - 1;
    }
  else
    g_assert_not_reached ();

  return ipuz_crossword_get_cell (xword, mirror_coord);
}

/* Calculates the apparent symmetry of the current grid */
static IPuzSymmetry
ipuz_crossword_calc_symmetry (IPuzCrossword *xword)
{
  guint row, col;
  gboolean is_quarter_symmetric = TRUE;

  /* Quarter-symmetric only makes sense if we're square */
  if (xword->width != xword->height)
    is_quarter_symmetric = FALSE;

  g_return_val_if_fail (IPUZ_IS_CROSSWORD (xword), IPUZ_SYMMETRY_NONE);

  for (row = 0; row < (((guint) xword->height + 1) / 2); row++)
    {
      for (col = 0; col < (guint) xword->width; col++)
        {
          IPuzCellCoord coord = { .row = row, .column = col };
          IPuzCell *cell;
          IPuzCell *mirror;

          cell = ipuz_crossword_get_cell (xword, coord);
          mirror = ipuz_crossword_cell_symmetry (xword, coord, 0);

          if (ipuz_cell_get_cell_type (cell) != ipuz_cell_get_cell_type (mirror))
            return IPUZ_SYMMETRY_NONE;

          if (is_quarter_symmetric)
            {
              mirror = ipuz_crossword_cell_symmetry (xword, coord, -1);
              if (ipuz_cell_get_cell_type (cell) != ipuz_cell_get_cell_type (mirror))
                is_quarter_symmetric = FALSE;

              mirror = ipuz_crossword_cell_symmetry (xword, coord, 1);
              if (ipuz_cell_get_cell_type (cell) != ipuz_cell_get_cell_type (mirror))
                is_quarter_symmetric = FALSE;
            }
        }
    }

  if (is_quarter_symmetric)
    return IPUZ_SYMMETRY_QUARTER;

  return IPUZ_SYMMETRY_HALF;
}

void
ipuz_crossword_set_symmetry (IPuzCrossword *xword,
                             IPuzSymmetry   symmetry)
{
  g_return_if_fail (IPUZ_IS_CROSSWORD (xword));

  /* QUARTER symmetry only makes sense on square boards */
  if (symmetry == IPUZ_SYMMETRY_QUARTER)
    g_return_if_fail (xword->width == xword->height);
  xword->symmetry = symmetry;

  /* FIXME(symmetry): we need to enforce */
}

/**
 * ipuz_crossword_get_symmetry:
 * @xword: A `IPuzCrossword`
 *
 * Returns the symmetry of @xword.
 *
 * Returns: The apparent symmetry of the puzzle.
 **/
IPuzSymmetry
ipuz_crossword_get_symmetry (IPuzCrossword *xword)
{
  g_return_val_if_fail (IPUZ_IS_CROSSWORD (xword), IPUZ_SYMMETRY_NONE);

  return xword->symmetry;
}

/**
 * ipuz_crossword_set_cell_type:
 * @xword: A `IPuzCrossword`
 * @coord: Coordinate of the cell to change
 * @cell_type: the new cell_type
 *
 * Sets the cell type at @coord to be @cell_type. This will respect
 * the symmetry of the puzzle, and update the other cells to have a
 * matching cell type
 **/
void
ipuz_crossword_set_cell_type (IPuzCrossword    *xword,
                              IPuzCellCoord     coord,
                              IPuzCellCellType  cell_type)
{
  IPuzCell *cell, *mirror;

  g_return_if_fail (IPUZ_IS_CROSSWORD (xword));

  cell = ipuz_crossword_get_cell (xword, coord);
  g_return_if_fail (cell != NULL);

  ipuz_cell_set_cell_type (cell, cell_type);

  if (xword->symmetry == IPUZ_SYMMETRY_HALF ||
      xword->symmetry == IPUZ_SYMMETRY_QUARTER)
    {
      mirror = ipuz_crossword_cell_symmetry (xword, coord, 0);
      ipuz_cell_set_cell_type (mirror, cell_type);
    }

  if (xword->symmetry == IPUZ_SYMMETRY_QUARTER)
    {
      mirror = ipuz_crossword_cell_symmetry (xword, coord, 1);
      ipuz_cell_set_cell_type (mirror, cell_type);
      mirror = ipuz_crossword_cell_symmetry (xword, coord, -1);
      ipuz_cell_set_cell_type (mirror, cell_type);
    }
}





static void
print_clues (GArray   *clues,
             gboolean  showenumerations)
{
  guint i;

  for (i = 0; i < clues->len; i++)
    {
      IPuzClue *clue = g_array_index (clues, IPuzClue *, i);
      g_autoptr (IPuzEnumeration) enumeration = NULL;

      enumeration = ipuz_clue_get_enumeration (clue);
      g_print ("\t");
      if (ipuz_clue_get_number (clue) > 0)
        g_print ("%d. ", ipuz_clue_get_number (clue));
      else if (ipuz_clue_get_label (clue))
        g_print ("%s. ", ipuz_clue_get_label (clue));
      if (ipuz_clue_get_clue_text (clue))
        g_print ("%s ", ipuz_clue_get_clue_text (clue));
      if (showenumerations && enumeration)
        g_print ("(%s)", ipuz_enumeration_get_display (enumeration));
      g_print ("\n");

      if (clue->cells->len > 0)
        {
          g_print ("\tcells: ");
          for (guint j = 0; j < clue->cells->len; j++)
            {
              IPuzCellCoord *coord = & (g_array_index (clue->cells, IPuzCellCoord, j));
              g_print ("[%u, %u] ", coord->row, coord->column);
            }
          g_print ("\n");
        }
    }
}

void
ipuz_crossword_print (IPuzCrossword *xword)
{
  g_return_if_fail (IPUZ_CROSSWORD (xword));

  /* Puzzle */
  g_autofree gchar *copyright = NULL;
  g_autofree gchar *publisher = NULL;
  g_autofree gchar *publication = NULL;
  g_autofree gchar *url = NULL;
  g_autofree gchar *uniquieid = NULL;
  g_autofree gchar *title = NULL;
  g_autofree gchar *intro = NULL;
  g_autofree gchar *explanation = NULL;
  g_autofree gchar *annotation = NULL;
  g_autofree gchar *author = NULL;
  g_autofree gchar *editor = NULL;
  g_autofree gchar *date = NULL;
  g_autofree gchar *notes = NULL;
  g_autofree gchar *difficulty = NULL;
  g_autofree gchar *charset = NULL;
  g_autofree gchar *origin = NULL;
  g_autofree gchar *block = NULL;
  g_autofree gchar *empty = NULL;

  /* Crossword */
  gint width;
  gint height;
  gboolean showenumerations;
  IPuzBoard *board;

  char ESC=27;


  g_object_get (G_OBJECT (xword),
                "title", &title,
                "copyright", &copyright,
                "publisher", &publisher,
                "publication", &publication,
                "url", &url,
                "uniqueid", &uniquieid,
                "intro", &intro,
                "explanation", &explanation,
                "annotation", &annotation,
                "author", &author,
                "editor", &editor,
                "date", &date,
                "notes", &notes,
                "difficulty", &difficulty,
                "charset", &charset,
                "origin", &origin,
                "block", &block,
                "empty", &empty,
                "showenumerations", &showenumerations,
                "width", &width,
                "height", &height,
                "board", &board,
                NULL);

  g_print ("\n");
  /* Print the title section */
  if (title) g_print ("%c[1m%s%c[0m\n", ESC, title, ESC);
  if (author || editor)
    {
      if (author) g_print ("\tby %s\t", author);
      if (editor) g_print ("\tedited by %s", editor);
      g_print ("\n");
    }
  if (copyright) g_print ("\tCopyright: %s\n", copyright);
  if (title || author || editor || copyright || date) g_print ("\n");

  for (int i = 0; i < width + 1; i++)
    g_print ("██");
  g_print ("\n");
  for (int row = 0; row < height; row ++)
    {
      g_print ("█");
      for (int column = 0; column < width; column ++)
        {
          IPuzCellCoord coord = { .row = row, .column = column };
          IPuzCell *cell = ipuz_board_get_cell (board, coord);
          gint n;
          switch (ipuz_cell_get_cell_type (cell))
            {
            case IPUZ_CELL_BLOCK:
              g_print ("▓▓");
              break;
            case IPUZ_CELL_NULL:
              g_print ("▞▚");
              break;
            case IPUZ_CELL_NORMAL:
              n = ipuz_cell_get_number (cell);
              if (n == 0)
                g_print ("  ");
              else
                g_print ((n<10?"%d ":"%d"),n);
              break;
            }
        }
      g_print ("█\n█");
      for (int column = 0; column < width; column ++)
        {
          const gchar *solution = NULL;
          IPuzCellCoord coord = { .row = row, .column = column };
          IPuzCell *cell = ipuz_board_get_cell (board, coord);
          switch (ipuz_cell_get_cell_type (cell))
            {
            case IPUZ_CELL_BLOCK:
              g_print ("▓▓");
              break;
            case IPUZ_CELL_NULL:
              g_print ("▚▞");
              break;
            case IPUZ_CELL_NORMAL:
               solution = ipuz_cell_get_solution (cell);
              if (solution)
                g_print (" %s", solution);
              else
                g_print ("  ");
              break;
            }
        }
      g_print ("█\n");
    }
  for (int column = 0; column < width + 1; column++)
    g_print ("██");
  g_print ("\n\n");

  GArray *clues;
  g_print ("%c[1mClues%c[0m\n", ESC, ESC);
  clues = ipuz_crossword_get_clues (xword, IPUZ_CLUE_DIRECTION_ACROSS);
  g_print ("\tAcross:\n");
  print_clues (clues, showenumerations);
  clues = ipuz_crossword_get_clues (xword, IPUZ_CLUE_DIRECTION_DOWN);
  g_print ("\n");
  g_print ("\tDown:\n");
  print_clues (clues, showenumerations);
  g_print ("\n");

  g_print ("%c[1mDocument Information%c[0m\n", ESC, ESC);
  if (date) g_print ("\tDate: %s\n", date);
  if (publisher) g_print ("\tPublisher: %s\n", publisher);
  if (publication) g_print ("\tPublication: %s\n", publication);
  if (url) g_print ("\tURL: %s\n:", url);
  if (uniquieid) g_print ("\tUnique ID:%s\n", uniquieid);
  if (difficulty) g_print ("\tDifficulty:%s\n", difficulty);

  g_print ("\n");

  g_print ("%c[1mDisplay Information%c[0m\n", ESC, ESC);
  if (block) g_print ("\tBlock string: '%s'\n", block);
  if (empty) g_print ("\tEmpty cell string: '%s'\n", empty);
  if (charset) g_print ("\tValid charset: '%s'\n", charset);
  g_print ("\tShow enumerations: %s\n", showenumerations?"true":"false");

  g_print ("\n");
#if 0
  "intro", &intro,
  "explanation", &explanation,
  "annotation", &annotation,
  "origin", &origin,
#endif


  g_object_unref (board);
}
